 /* The test application, described in the documentation of UthLib */
#include <pthread.h>
#include "uth.h"
#include <stdio.h>

#define NumThreads 2

uth_t worker[NumThreads];
uth_t virtualprocessorA;
uth_t virtualprocessorB;

int global_var = 0;

void workerFunc(void* arg)
{
	int id = (int)arg;
	int local_var = 0;

	global_var++;
	local_var++;

	printf("[0x%lx] round one: arg (%d / %d)\t [local = %d global = %d]\n",
		uth_self(), id, (int) uth_getarg(uth_self()), local_var, global_var);

	if (id == NumThreads-1)
		uth_switchto2(0, worker[id], virtualprocessorA);
	else
		uth_switchto2(0, worker[id], worker[id+1]);

	global_var++;
	local_var++;

	printf("[0x%lx] round two: arg (%d / %d)\t [local = %d global = %d]\n",
		uth_self(), id, (int) uth_getarg(uth_self()), local_var, global_var);

	if (id == NumThreads-1)
		uth_switchto2(1, NULL, virtualprocessorB);
	else
		uth_switchto2(1, NULL, worker[id+1]);
}

void *kernelthreadfunc(void *arg)
{
	uth_vp_init(1);
	virtualprocessorB = uth_self();

	printf("[0x%lx] master thread B starts\t [pthread_t = 0x%lx]\n",
		uth_self(), pthread_self());

	uth_switchto2(1, virtualprocessorB, worker[0]);
	printf("[0x%lx] master thread B exits\t [global = %d]\n",
		uth_self(), global_var);

	return 0;
}

int main(void)
{
	int i;
	pthread_t pth;
	pthread_attr_t attr;

	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	uth_init(0);
	uth_vp_init(0);
	for(i=0; i<NumThreads; ++i){
		worker[i] = uth_create(workerFunc,(void *)i);
	}

	printf("[0x%lx] master thread A starts\t [pthread_t = 0x%lx]\n",
		uth_self(), pthread_self());
	virtualprocessorA = uth_self();
	uth_switchto2(0, virtualprocessorA, worker[0]);
	printf("[0x%lx] master thread A continues\t [global = %d]\n",
		uth_self(), global_var);

/**/	pthread_create(&pth, &attr, kernelthreadfunc, NULL);
	pthread_join(pth, NULL);/**/

	for(i=0; i<NumThreads; ++i){
		uth_delete(worker[i]);
	}

	printf("[0x%lx] master thread A exits\n", uth_self());
	return 0;
}