/* 
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the Netscape Portable Runtime library.
 * 
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are 
 * Copyright (C) 1994-2000 Netscape Communications Corporation.  All
 * Rights Reserved.
 * 
 * Contributor(s):  Silicon Graphics, Inc.
 * 
 * Portions created by SGI are Copyright (C) 2000-2001 Silicon
 * Graphics, Inc.  All Rights Reserved.
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the MPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the MPL or the
 * GPL.
 */

/*
 * This file is derived directly from the State Threads library, 
 * modified extensively by Panagiotis E. Hadjidoukas for UthLib. 
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#include <fcntl.h>
#else
#define _WIN32_WINNT	0x0500
#include <windows.h>
#endif
#include <string.h>
#include "common.h"

#ifdef REUSE_THREADS

void _uth_queue_init (register struct uth_single_queue * q)
{
	q->first = NULL;
	pthread_mutex_init(&q->mutex, NULL);
}

void _uth_to_reuseq (register uth_t *nth)
{
    pthread_mutex_lock (&reuse_uth_q.mutex);
    if (uth_queue_empty (&reuse_uth_q))
        reuse_uth_q.last = nth;
    nth->next = reuse_uth_q.first;
    reuse_uth_q.first = nth;
    pthread_mutex_unlock (&reuse_uth_q.mutex);
}

uth_t *_uth_reuseq_dequeue ()
{
	register uth_t *nth;
    
	nth = reuse_uth_q.first;
	if (nth==NULL) return NULL;

	pthread_mutex_lock (&reuse_uth_q.mutex);
	nth = reuse_uth_q.first;
	if (nth!=NULL) {
		reuse_uth_q.first = nth->next;
		nth->next = NULL;
	}
	pthread_mutex_unlock (&reuse_uth_q.mutex);
	return nth;
}

void _uth_to_reuselrq (int which_q, register uth_t *nth)
{
    pthread_mutex_lock (&reuse_uth_lq.q[which_q].mutex);
    if (uth_queue_empty (&reuse_uth_lq.q[which_q]))
        reuse_uth_lq.q[which_q].last = nth;
    nth->next = reuse_uth_lq.q[which_q].first;
    reuse_uth_lq.q[which_q].first = nth;
    pthread_mutex_unlock (&reuse_uth_lq.q[which_q].mutex);
}

uth_t *_uth_reuselrq_dequeue (int which_q)
{
    register uth_t *nth;

    nth = reuse_uth_lq.q[which_q].first;
    if (nth==NULL) return NULL;

    pthread_mutex_lock (&reuse_uth_lq.q[which_q].mutex);
    nth = reuse_uth_lq.q[which_q].first;
    if (nth!=NULL) {
        reuse_uth_lq.q[which_q].first = nth->next;
        nth->next = NULL;
    }
    pthread_mutex_unlock (&reuse_uth_lq.q[which_q].mutex);

    return nth;
}

uth_t *_uth_get_reused_thread(void)
{
	uth_t *thread;

#ifdef LOCAL_REUSE_QUEUES
	int vp = uth_get_vpid();
	thread = _uth_reuselrq_dequeue (vp);
#else
	thread = _uth_reuseq_dequeue();
#endif

	return thread;
}

void _uth_put_reused_thread(uth_t *thread)
{
#ifdef LOCAL_REUSE_QUEUES
	int vp = uth_get_vpid();

	_uth_to_reuselrq(vp, thread);
#else
	_uth_to_reuseq(thread);
#endif
}
#endif


/* Global data */
uth_data_t		_uth_data;            /*  */

#if defined(WIN32)
size_t getpagesize(void)
{
	SYSTEM_INFO sysinf;
	GetSystemInfo(&sysinf);
	return (sysinf.dwPageSize);
}
#elif defined(QNX)
size_t getpagesize(void)
{
	return 4096;
}
#endif

int uth_init(int stacksize)
{
  static int initialized = 0;

  if (!initialized){
#ifdef REUSE_THREADS
	  int i;

  	  memset(&_uth_data, 0, sizeof(uth_data_t));

	  _uth_queue_init(&reuse_uth_q);
	  for (i = 0; i < UTH_MAX_CPUS; i++) {
		  _uth_queue_init(&reuse_uth_lq.q[i]);
	  }
#endif
	  _uth_data.pagesize = getpagesize();
	  if (stacksize == 0)
		  _uth_data.stacksize = UTH_DEFAULT_STACK_SIZE;
	  else
		  _uth_data.stacksize = stacksize;
	  initialized = 1;
  }
  return 0;
}

int uth_vp_init(int vp)
{
  uth_t *thread;
  static int initialized = 0;

  if (!initialized){
#ifdef REUSE_THREADS
	  int i;

  	  memset(&_uth_data, 0, sizeof(uth_data_t));

	  _uth_queue_init(&reuse_uth_q);
	  for (i = 0; i < UTH_MAX_CPUS; i++) {
		  _uth_queue_init(&reuse_uth_lq.q[i]);
	  }
#endif
	  _uth_data.pagesize = getpagesize();
	  _uth_data.stacksize = UTH_DEFAULT_STACK_SIZE;
	  initialized = 1;
  }

  vp_ids[vp][0] = pthread_self();	/* store vpid */

  thread = (uth_t *) calloc(1, sizeof(uth_t));
  if (!thread)
    return -1;

  _UTH_SET_CURRENT_THREAD(vp, thread);

  return 0;
}

int uth_get_vpid(void)
{
	int i;

	pthread_t me = pthread_self();

	for (i = 0; i < UTH_MAX_CPUS; i++)
		if (pthread_equal(me, vp_ids[i][0])) return i;
		
	printf("fatal error\n");
	exit(1);

	/* Not reached */
	return -1;
}


void _uth_main(void)
{ 
	uth_t *thread;
	int vp  = uth_get_vpid();
	thread = _UTH_CURRENT_THREAD(vp);

	/* Run thread main */
	(*thread->start)(thread->arg);
}

uth_t *uth_create(void (*start)(void *arg), void *arg)
{
	uth_t *thread;
	uth_stack_t *stack;
	uth_stack_t mystack;
	int stk_size;

	char *sp;
#ifdef __ia64__
	char *bsp;
#endif

	stk_size = _UTH_STACK_SIZE;
#ifdef REUSE_THREADS
	if ((thread = _uth_get_reused_thread()) == NULL) {
#endif
		/* Adjust stack size */
		stk_size = ((stk_size + _UTH_PAGE_SIZE - 1) / _UTH_PAGE_SIZE) * _UTH_PAGE_SIZE;
		stack = _uth_stack_new(&mystack, stk_size);
		if (!stack)
			return NULL;

		/* Allocate thread object and per-thread data off the stack */
#if defined (MD_STACK_GROWS_DOWN)
		sp = stack->stk_top;
#ifdef __ia64__
		/*
		 * The stack segment is split in the middle. The upper half is used
		 * as backing store for the register stack which grows upward.
		 * The lower half is used for the traditional memory stack which
		 * grows downward. Both stacks start in the middle and grow outward
		 * from each other.
		 */
		sp -= (stk_size >> 1);
		bsp = sp;
		/* Make register stack 64-byte aligned */
		if ((unsigned long)bsp & 0x3f)
			bsp = bsp + (0x40 - ((unsigned long)bsp & 0x3f));
		stack->bsp = bsp + _UTH_STACK_PAD_SIZE;
#endif
		sp = sp - sizeof(uth_t);
		thread = (uth_t *) sp;

		/* Make stack 64-byte aligned */
		if ((unsigned long)sp & 0x3f)
			sp = sp - ((unsigned long)sp & 0x3f);
		stack->sp = sp - _UTH_STACK_PAD_SIZE;
#elif defined (MD_STACK_GROWS_UP)
		sp = stack->stk_bottom;
		thread = (uth_t *) sp;
		sp = sp + sizeof(uth_t);

		/* Make stack 64-byte aligned */
		if ((unsigned long)sp & 0x3f)
			sp = sp + (0x40 - ((unsigned long)sp & 0x3f));
		stack->sp = sp + _UTH_STACK_PAD_SIZE;
#else
#error Unknown OS
#endif
		memset(thread, 0, sizeof(uth_t));

		/* Initialize thread */
		memcpy(&thread->stackinfo, stack, sizeof(uth_stack_t));
		thread->stack = &thread->stackinfo;

		thread->start = start;
		thread->arg = arg;
		thread->retval = 0xDEAD;

#ifndef __ia64__
		_UTH_INIT_CONTEXT(thread, stack->sp, _uth_main);
#else
		_UTH_INIT_CONTEXT(thread, stack->sp, stack->bsp, _uth_main);
#endif

#ifdef REUSE_THREADS
	}
	else {
		uth_reinit(thread, start, arg);
	}
#endif

	return thread;
}

void uth_reinit(uth_t *thread, void (*start)(void *arg), void *arg)
{
	thread->start = start;
	thread->arg = arg;
#ifndef __ia64__
	_UTH_INIT_CONTEXT(thread, thread->stack->sp, _uth_main);
#else
	_UTH_INIT_CONTEXT(thread, thread->stack->sp, thread->stack->bsp, _uth_main);
#endif
}

void uth_delete(uth_t *thread)
{
#ifdef REUSE_THREADS
	_uth_put_reused_thread(thread);
#else
	_uth_stack_free(thread->stack);
#endif
}

void uth_switchto(uth_t *old, uth_t *next)
{
	int vp = uth_get_vpid();

#if (CTXSW_METHOD == CTXSW_SJLJ)
#ifdef WIN32	/* This is only for windows */
	unsigned long regs[6];
#endif
#endif

	if (next == NULL) {
		printf("uth_switchto: next == NULL\n");
		exit(1);
	}

#if (CTXSW_METHOD == CTXSW_SJLJ)
#ifdef WIN32	/* This is only for windows */
	memcpy(regs, &next->context, sizeof(regs));
	MD_SETJMP(next->context);	/* re-init */
	memcpy(&next->context, regs, sizeof(regs));
#endif
#endif

#if (CTXSW_METHOD == CTXSW_SJLJ)
#ifdef WIN32	/* This is only for windows */
	memcpy(regs, &next->context, sizeof(regs));
	MD_SETJMP(next->context);	/* re-init */
	memcpy(&next->context, regs, sizeof(regs));
#endif
#endif


	if (old != NULL) {
		_UTH_SWITCH_CONTEXT(vp, old, next);
	}
	else {
		_UTH_RESTORE_CONTEXT(vp, next);
	}
}

uth_t *uth_self()
{
	int vp = uth_get_vpid();

	return _UTH_CURRENT_THREAD(vp);
}

void *uth_getarg(uth_t *thread)
{
	return thread->arg;
}

void uth_setarg(uth_t *thread, void *arg)
{
	thread->arg = arg;
}


/* Secondary (Extended API) */
uth_t *uth_self2(int vp)
{
	return _UTH_CURRENT_THREAD(vp);
}


void uth_switchto2(int vp, uth_t *old, uth_t *next)
{
#if (CTXSW_METHOD == CTXSW_SJLJ)
#ifdef WIN32	/* This is only for windows */
	unsigned long regs[6];
#endif
#endif

	if (next == NULL) {
		printf("uth_switchto: next == NULL\n");
		exit(1);
	}

#if (CTXSW_METHOD == CTXSW_SJLJ)
#ifdef WIN32	/* This is only for windows */
	memcpy(regs, &next->context, sizeof(regs));
	MD_SETJMP(next->context);	/* re-init */
	memcpy(&next->context, regs, sizeof(regs));
#endif
#endif

	if (old != NULL) {
		_UTH_SWITCH_CONTEXT(vp, old, next);
	}
	else {
		_UTH_RESTORE_CONTEXT(vp, next);
	}
}

